function Results = CHL_Algorithm(y,X,firm_count,b_NR,num_iter,jones,smooth)

% Inputs:   y           :   The full vector of return data stacked by cross-section then month
%           X           :   Explanatory variable (Obs x K) stacked to correspond with y
%	        gamma	    :   time series of root-mean-square asset specific return
%           firm_count  :   Contains for each period the number of observations
%                           (is used to collect right y,X in each period) - note
% 			                it has repeated constant values over 12-month periods
%           b_NR        :   Connor-Linton bandwidth (0.05 e.g. means 5%
%                           of the nearest data points are within one bandwidth)
%           num_iter    :   Number of iterations
%           jones       :   0 for unweighted, 1 for weighting  la Jones (2001)
%           smooth      :   0 for local constant smoother, 1 for local linear
%
%           It is assumed that the Xs are standardized
%           
%
% Outputs:  FactRets    :   Factor Returns (first column market exposure = constant)
%           Char_Func   :   Characteristic functions
%           Unc_R2_lin  :   Uncentered R2 linear model
%           Unc_R2_chl  :   Uncentered R2 nonlinear model
%
% To run the code: Results = CHL_Algorithm(y,X,num_firms,7,5,0,0);
%           
% ====================================================================================
% set defaults for weight and smooth
if nargin<7
   smooth = 1; % local linear estimator is the default
end
if nargin<6
   weight = 0; % no weighting is the default
end
num_periods = length(firm_count);
mean_num = mean(num_periods);
cum_firms = cumsum(firm_count); % cumulative number of firms
J = size(X,2); % The number of characteristic-based factors
Tot_Num = length(y); % Total number of observations across all periods
% ===========================================================================================
%                      Initialization and Bandwidth Determination
% ===========================================================================================
% Intialize the characteristic functions to be linear

cb = (-3:0.1:3)'; % the grid from -3 to 3 in .1 intervals for target points

G = repmat(cb,1,J); 
disp('Initialization of the Model and construction of local bandwidth');

% Initialize the factor returns based on linear characteristic functions
% and get an estimate for Jones (2001) weighting term.

% loop over all time periods
for period_loop = 1:num_periods
    % start and end index of the data 
    % inside loop runs from sum of number of firms over all time periods up to last period
    % to sum of number of firms over all time periods including this one
    start_index(period_loop) = [1 + cum_firms(period_loop) - firm_count(period_loop)];
    end_index(period_loop) = cum_firms(period_loop);
    % For each characteristic compute at each grid point a bandwidth taking into account location as well as sample size.
    % (collect in bandwidth matrix B (grid x time x J)
    
    for j=1:J % loop over the number of characteristic-based factors
        X_help = X(start_index(period_loop):end_index(period_loop),j);
        for g = 1:length(cb)
            Y = sqrt((X_help - cb(g)).^2);
            B(g,period_loop,j) = prctile(Y,b_NR)*(firm_count(period_loop)/mean_num)^-0.2;
        end 
% end of loop over the number of characteristics    	
	end 
% end of loop over cross-section of firms

    % Compute linear factor returns for initialization
    FactRets(period_loop,:) = speedols(y(start_index(period_loop):end_index(period_loop)),[ones(firm_count(period_loop),1) X(start_index(period_loop):end_index(period_loop),:)])';
    
    switch jones
        case 0 
% unweighted case
             gamma(period_loop,1) = 1;
        case 1 
% weighted case: variance of residuals in this period
             fit_lin = sum(([ones(firm_count(period_loop),1) X(start_index(period_loop):end_index(period_loop),:)]*FactRets(period_loop,:)),2);
             eps_lin = y(start_index(period_loop):end_index(period_loop)) - fitlin;
             gamma(period_loop,1) = var(eps_lin);
    end
             
end 
% end of loop over time periods

Results.Linear_FactRets = FactRets;

disp('Initialization of the model finished, proceed to Iteration');

% ===========================================================================================
%                                       Iteration
% =========================================================================================== 
for iter = 1:num_iter 
% loop over the number of iterations for convergence

	% Knowing the factor returns we estimate the characteristic functions
    % ===================================================================
	for j = 1:J 
% loop over characteristic-based factors
        disp(['Iteration ' num2str(iter) ' in Characteristic Function Estimation, Factor ' num2str(j)]);
        for period_loop = 1:num_periods
            % based on linear approximation determine the G matrix over the full required grid and multiply by factor returns
            for j_2=1:J
                G_Full(:,j_2) = interp1(cb, G(:,j_2), X(start_index(period_loop):end_index(period_loop),j_2));
            end
            % multiply by factor returns and delete out the jth component
            G_Full_fact = G_Full.*repmat(FactRets(period_loop, 2:end),size(G_Full,1),1);
            G_Full_fact(:,j) = [];
            % determine the partial residual
            y_part = y(start_index(period_loop):end_index(period_loop)) - FactRets(period_loop,1) - sum(G_Full_fact,2);
            % Do local nonparametric smoothing at each required grid point
            for g = 1:length(cb); 
% loop over the set of target points
                % generate the kernel weights assuming h(g) contains grid point specific bandwidth
                K=(X(start_index(period_loop):end_index(period_loop),j)-cb(g))/B(g,period_loop,j);
                switch smooth
                    case 0 
% local constant smoother
                         W=(feval(@kgaussian,K));
                         lambda(period_loop,g) = (y_part'*W)/sum(W);
                    case 1 
% local linear smoother
                         W=(feval(@kgaussian,K)).^0.5;
                         Y_t=W.*y_part;
                         X_t=repmat(W,1,2).*[ones(firm_count(period_loop),1) (X(start_index(period_loop):end_index(period_loop),j)-cb(g))];
                         b = speedols(Y_t,X_t);
                         lambda(period_loop,g) = b(1); % the required nonparametric estimate at grid point g
                 end
            end 
% end of the grid loop
            clear G_Full_fact G_Full;
        end 
% end of the time-period loop

        % We have now all inputs ready to update the characteristic functions at all grid points
        % Compute normalization (sum of squared fact rets)
        help_FactRets = FactRets(:,j+1); 
% Vector containing all factor returns to char j
	    wtd_FactRets = help_FactRets./gamma; 
% factors weighted them by the reciprocal of gamma
        norm_FactRets = sum(wtd_FactRets.*help_FactRets);
        for g = 1:length(cb)
            help_lambdas = lambda(:,g);
            G(g,j) = sum(wtd_FactRets.*help_lambdas)/norm_FactRets;
        end
        % Normalize the G functions as in the paper, on the grid points only to avoid computation over the full cross section, we
        % do this at the end.
        G(:,j) = (G(:,j) - mean(G(:,j)))/std(G(:,j));
        G_iter(:,j,iter) = G(:,j);
    end
	% Having an update of the characteristic functions, reestimate the factor returns
    % ===============================================================================
    disp(['Iteration ' num2str(iter) ' in Factor Return Updating']);

    for period_loop = 1:num_periods
        % Compute the characteristic functions at all relevant values in this period_loop
        for j_3=1:J
            X_full(:,j_3) = interp1(cb,G(:,j_3), X(start_index(period_loop):end_index(period_loop),j_3));
        end
        FactRets(period_loop,:) = speedols(y(start_index(period_loop):end_index(period_loop)),[ones(firm_count(period_loop),1) X_full])';
        clear X_full;
    end 
% end of the period loop
end 
% end of iteration to convergence

Results.G = G;
Results.G_iter = G_iter;
Results.Nonlinear_FactRets = FactRets;
Results.Bandwidth = B;

% ===========================================================================================
%  Diagnostics/Standard Errors for G functions as well as t-stats for period Factor Returns
% ===========================================================================================

for period_loop = 1:num_periods
    % Prepare linear and "nonlinear" explanatory matrices
    X_lin = X(start_index(period_loop):end_index(period_loop),:);
    for j_2=1:J
        X_nlin(:,j_2) = interp1(cb, G(:,j_2), X(start_index(period_loop):end_index(period_loop),j_2));
    end 
% end of time period loop

    % Compute ols regression of return on nonlinear explanatory variables to obtain hetero-consistent t-stats for factor returns
    [b,bint,r,rint,stats,tstats] = dohwregress(y(start_index(period_loop):end_index(period_loop)),[ones(firm_count(period_loop),1) X_nlin]);
    Results.T_statistics(period_loop,:) = tstats';

    % Compute uncentered R2 of the linear/nonlinear model
    fit_lin = Results.Linear_FactRets(period_loop, 1) + sum(X_lin.*repmat(Results.Linear_FactRets(period_loop, 2:end),size(X_lin,1),1),2);
    fit_nlin = Results.Nonlinear_FactRets(period_loop, 1) + sum(X_nlin.*repmat(Results.Nonlinear_FactRets(period_loop, 2:end),size(X_nlin,1),1),2);

    % Compute one factor (constant) MSE of the model in cross section
    eps_bench = y(start_index(period_loop):end_index(period_loop))-mean(y(start_index(period_loop):end_index(period_loop)));
    Results.UR2_Bench(period_loop,1) = 1 - sum(eps_bench.^2)/sum(y(start_index(period_loop):end_index(period_loop)).^2);
    eps_linear = y(start_index(period_loop):end_index(period_loop)) - fit_lin;
    Results.UR2_Linear(period_loop,1) = 1 - sum(eps_linear.^2)/sum(y(start_index(period_loop):end_index(period_loop)).^2);
    eps_nonlinear = y(start_index(period_loop):end_index(period_loop)) - fit_nlin;
    Results.UR2_NonLinear(period_loop,1) = 1 - sum(eps_nonlinear.^2)/sum(y(start_index(period_loop):end_index(period_loop)).^2);    
    % Compute Residual Variance for Standard Errors
    resid_var(period_loop,1) = var(y(start_index(period_loop):end_index(period_loop),1) - fit_nlin);
    % And collect it as a result output
    Results.Resid_MSE(period_loop,1) = std(eps_nonlinear);
    % =======================================================
    % Compute the Mammen & Park cross validation criterion
    % =======================================================
    B_help = squeeze(B(:,period_loop,:)); 
% This is a grid x J matrix containing bandwidths used at grid points in period "period_loop".
    % Get bandwidths at each single observation point to evaluate the Mammen & Park criterion
    for j_b=1:J
        B_app(:,j_b) = interp1(cb, B_help(:,j_b), X(start_index(period_loop):end_index(period_loop),j_b)); 
% n(t) x J matrix
    end
    % Compute vector containing (1 + 2K(0) + sum((1/nT)*h(i,j,t)))
    mamm_help = (1+2*kgaussian(0)*(1/Tot_Num)*sum((1./B_app),2));
    MP(period_loop,1) = sum((eps_nonlinear.^2).*mamm_help);    
    clear X_lin X_nlin fit_lin fit_nlin eps_bench eps_linear eps_nonlinear B_help B_app;  
end 
% end period loop

% Compute the full Mammen & Park Criterion over the full sample period
Results.MP = sum(MP); 
% do not divide as otherwise not enough number after the comma...

% Compute the standard errors at each grid point for each factor j
% Proxy for the density value at each grid point (assume normality)
dens_prox = normpdf(cb,0,1);
for j=1:J
    help_FactRets = Results.Nonlinear_FactRets(:,j+1);
    norm_FactRets = mean(help_FactRets.^2); % mean of squared factor returns
    W_Resid_Var = mean((help_FactRets.^2).*resid_var);
    for g = 1:length(cb)
        % Calculate the mean bandwidth used at this grid point
        mean_band = mean(squeeze(B(g,:,j)));
        Results.STD(g,j) = ((1/(size(X,1)*mean_band))*0.282*(W_Resid_Var/(dens_prox(g)*(norm_FactRets.^2)))).^0.5;
    end
end
% To be consistent with renormalization in the paper:
% First save the results obtained from zero one standardization
Results.G_zero_one_std = Results.G;
Results.Nonlinear_FactRets_zero_one_std = Results.Nonlinear_FactRets;
Results.STD_zero_one_std = Results.STD;
for j=1:J
    % Compute G at every single sample point
    G_super(:,j) = interp1(cb, G(:,j), X(:,j));
    sigma_std(j) = std(G_super(:,j));
    mean_std(j) = mean(G_super(:,j));
    % correct G functions
    Results.G(:,j) = (Results.G(:,j)-mean_std(j))/sigma_std(j);
    % correct intercept
    Results.Nonlinear_FactRets(:,1) = Results.Nonlinear_FactRets(:,1) + mean_std(j)*Results.Nonlinear_FactRets(:,j+1);
    % rescale Factor Returns
    Results.Nonlinear_FactRets(:,j+1) = Results.Nonlinear_FactRets(:,j+1)*sigma_std(j);
    % rescale the standard errors of the g-functions (mean no effect)
    Results.STD(:,j) = Results.STD(:,j)/sigma_std(j);
end